local TIMELINE_MIN = FRACUNIT*5 //what is considered left of timeline
local TIMELINE_MAX = FRACUNIT*300 //what is considered right of timeline

rawset(_G,"chrono_timescale",CV_RegisterVar({
	name = "chrono_timescale",
	defaultvalue = "Normal",
	PossibleValue = {Stopped = 0, Quarter = 1, Half = 2, Normal = 3, Double = 4, Triple = 5},
	flags = CV_SHOWMODIF|CV_NOTINNET
}))

rawset(_G,"chrono_indicators",CV_RegisterVar({
	name = "chrono_indicators",
	defaultvalue = "On",
	PossibleValue = CV_OnOff,
	flags = CV_SHOWMODIF|CV_NETVAR
}))

rawset(_G,"chrono_blocks",CV_RegisterVar({
	name = "chrono_blocks",
	defaultvalue = "On",
	PossibleValue = CV_OnOff,
	flags = CV_SHOWMODIF|CV_NETVAR
}))

rawset(_G,"chrono_timeline",CV_RegisterVar({
	name = "chrono_timeline",
	defaultvalue = "On",
	PossibleValue = CV_OnOff,
	flags = CV_SHOWMODIF|CV_CALL,
	func = function(var)
		if var.value == 1
			hud.enable("lives")
		else
			hud.disable("lives")
		end
	end
}))


freeslot("SPR_CHBL","S_CHRONO_BLOCK","MT_CHRONOBLOCK")

states[S_CHRONO_BLOCK] = {
	sprite = SPR_CHBL,
	frame = A,
	duration = -1,
	next = S_CHRONO_BLOCK
}
mobjinfo[MT_CHRONOBLOCK] = {
	spawnstate = S_CHRONO_BLOCK,
	deathstate = S_NULL,
	spawnhealth = 1000,
	height = 35*FRACUNIT,
	radius = 16*FRACUNIT,
	flags = MF_SOLID|MF_SLIDEME
}



local function addToPlayerTimeline(player,time,eventInfo)
	if not player
		return false
	end
	
	if not eventInfo
		return false
	end
	
	if player.invalidatedIter
		return false
	end
	
	rawset(player.timeline,time,eventInfo)
	return true
end






addHook("ThinkFrame", function()
	if not chronoSoundlist
		rawset(_G,"chronoSoundlist",{1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,102,106,105,97,110,123,124,125,126,127,128,129,130,132,137})
	end

	for player in players.iterate do
		if player.mo
			if player.iterations == nil
				player.iterations = {}
			end
			
			if player.chrframe == nil
				player.chrframe = 1
			end
			
			if player.recframe == nil
				player.recframe = 1
			end
			
			if player.iteration == nil
				player.iteration = 1
			end
			
			if player.curTime == nil
				player.curTime = 1
			end
			
			if player.timeText == nil
				player.timeText = ""
			end
			
			if player.origMap == nil
				player.origMap = gamemap
			end
			
			if player.timeline == nil
				player.timeline = {
					[1] = {"T_START"},
					[TICRATE*60] = {"T_1MIN"},
					[TICRATE*120] = {"T_2MIN"},
					[TICRATE*180] = {"T_3MIN"},
					[TICRATE*240] = {"T_4MIN"},
					[TICRATE*300] = {"T_5MIN"},
					[TICRATE*360] = {"T_6MIN"},
					[TICRATE*420] = {"T_7MIN"},
					[TICRATE*480] = {"T_8MIN"},
					[TICRATE*540] = {"T_9MIN"},
					[TICRATE*600] = {"T_10MIN"},
				}
			end
			
			for eventTime,eventProp in pairs(player.timeline) do
				if player.curTime == tonumber(eventTime)
					if eventProp[1] == "T_START"
						player.timeText = "Started level"
					end
					if eventProp[1] == "T_ITERATETFLG"
						player.timeText = "Iteration " .. eventProp[2] .. " willingly iterated"
					end
					if eventProp[1] == "T_ITERATE"
						local deathCause
						if eventProp[3] == DMG_FIRE
							deathCause = "being burnt"
						elseif eventProp[3] == DMG_ELECTRIC
							deathCause = "being electrocuted"
						elseif eventProp[3] == DMG_WATER
							deathCause = "poisonous water"
						elseif eventProp[3] == DMG_SPIKE
							deathCause = "touching spikes"
						elseif eventProp[3] == DMG_NUKE
							deathCause = "being nuked"
						elseif eventProp[3] == DMG_INSTAKILL
							deathCause = "forces of instantaneous death"
						elseif eventProp[3] == DMG_SPACEDROWN
							deathCause = "suffocating in space"
						elseif eventProp[3] == DMG_DROWNED
							deathCause = "drowning"
						elseif eventProp[3] == DMG_CRUSHED
							deathCause = "being crushed"
						elseif eventProp[3] == DMG_DEATHPIT
							if eventProp[5] == true
								deathCause = "sinking in quicksand"
							else
								deathCause = "falling into a pit"
							end
						else
							deathCause = "generic causes"
						end
						player.timeText = "Iteration " .. eventProp[2] .. " died from " .. deathCause
					end
					if eventProp[1] == "T_CHECKPOINT"
						player.timeText = "Iteration " .. eventProp[2] .. " reached checkpoint " .. eventProp[3]+1
					end
					if eventProp[1] == "T_SUFFOCATE"
						player.timeText = "Iteration " .. eventProp[2] .. " is drowning"
					end
					if eventProp[1] == "T_BUBBLE"
						player.timeText = "Iteration " .. eventProp[2] .. " inhaled a bubble"
					end
					if eventProp[1] == "T_LEAVEWATER"
						player.timeText = "Iteration " .. eventProp[2] .. " left the water"
					end
					if eventProp[1] == "T_SUPER"
						player.timeText = "Iteration " .. eventProp[2] .. " turned Super"
					end
					if eventProp[1] == "T_NOSUPER"
						player.timeText = "Iteration " .. eventProp[2] .. " lost Super state"
					end
					if eventProp[1] == "T_1MIN"
						player.timeText = "One-minute mark"
					end
					if eventProp[1] == "T_2MIN"
						player.timeText = "Two-minute mark"
					end
					if eventProp[1] == "T_3MIN"
						player.timeText = "Three-minute mark"
					end
					if eventProp[1] == "T_4MIN"
						player.timeText = "Four-minute mark"
					end
					if eventProp[1] == "T_5MIN"
						player.timeText = "Five-minute mark"
					end
					if eventProp[1] == "T_6MIN"
						player.timeText = "Six-minute mark"
					end
					if eventProp[1] == "T_7MIN"
						player.timeText = "Seven-minute mark"
					end
					if eventProp[1] == "T_8MIN"
						player.timeText = "Eight-minute mark"
					end
					if eventProp[1] == "T_9MIN"
						player.timeText = "Nine-minute mark"
					end
					if eventProp[1] == "T_10MIN"
						player.timeText = "Ten-minute mark"
					end
				end
			end
			
			if player.viewingIter != nil
				local indicator = P_SpawnMobjFromMobj(player.mo,0,0,player.mo.height + (FRACUNIT*5),MT_THOK)
				indicator.fuse = 1
				indicator.sprite = SPR_LCKN
				indicator.frame = E|FF_FULLBRIGHT
			end
			
			if player.iterations and player.iteration and player.deadtimer < TICRATE/2
				if not player.iterations[player.iteration]
					player.iterations[player.iteration] = {}
				else
					local playingSounds = {}
					for sound = 1,#chronoSoundlist do
						if S_SoundPlaying(player.mo,chronoSoundlist[sound])
							local pastIter = player.iterations[player.iteration]
							if pastIter[player.recframe-1]
								//if wasSoundPlayed == false
								table.insert(playingSounds,chronoSoundlist[sound])
								//break
								//end
							end
						end
					end
					rawset(player.iterations[player.iteration],player.recframe,{player.mo.x,  player.mo.y,  player.mo.z,  player.mo.angle,  player.mo.skin,  player.mo.color,  player.mo.state,  player.mo.frame,  player.drawangle,  player.mo.rollangle,  player.mo.scale,  player.mo.colorized,  player.mo.flags,  player.mo.flags2,  player.mo.eflags,  player.mo.renderflags,  player.speed,  player.powers[pw_shield],  playingSounds, camera.x,camera.y,camera.z,camera.angle,camera.aiming,  player.starpostnum,  player.powers[pw_super],  player.mo.spriteyoffset,  player.mo.spritexoffset,  player.mo.spritexscale,  player.mo.spriteyscale,  player.mo.mirrored,  player.mo.blendmode,  player.mo.height,  player.mo.radius,  player.powers[pw_invulnerability]})
					player.recframe = $ + 1
				end
			end
			
			
			if player.powers[pw_underwater] == 387
				addToPlayerTimeline(player,player.curTime,{"T_SUFFOCATE",player.iteration})
				player.suffocatingInWater = true
			end
			
			if not (player.mo.eflags & MFE_UNDERWATER) and player.suffocatingInWater
				addToPlayerTimeline(player,player.curTime,{"T_LEAVEWATER",player.iteration})
				player.suffocatingInWater = nil
			end
			
			
			
			if player.powers[pw_super] and not player.hasalreadyturnedsuper
				addToPlayerTimeline(player,player.curTime,{"T_SUPER",player.iteration})
				player.hasalreadyturnedsuper = true
			end
			
			if not player.powers[pw_super] and player.hasalreadyturnedsuper
				addToPlayerTimeline(player,player.curTime,{"T_NOSUPER",player.iteration})
				player.hasalreadyturnedsuper = nil
			end
			
			
			//The actual nitty-gritty of getting all the ghosts to work n' stuff
			//I may have borrowed this method from Bloops's Tape Recorder
			if player.iterations
				for iteration,frames in pairs(player.iterations) do
					if iteration ~= player.iteration
						local curFrame = frames[player.chrframe]
						if curFrame
							local ghost = P_SpawnMobj(curFrame[1],curFrame[2],curFrame[3],MT_THOK)
							ghost.color = curFrame[6]
							ghost.skin = curFrame[5]
							ghost.fuse = 1
							ghost.sprite = SPR_PLAY
							ghost.angle = curFrame[9]
							//I don't know why but the system just refuses to capture super sprites X_X
							if curFrame[26] > 0
								ghost.state = S_PLAY_NIGHTS_FLOAT
								ghost.frame = A
							else
								ghost.state = curFrame[7]
								ghost.frame = curFrame[8]
							end
							ghost.rollangle = curFrame[10]
							ghost.scale = curFrame[11]
							ghost.colorized = curFrame[12]
							ghost.flags2 = curFrame[14]
							ghost.eflags = curFrame[15]
							if curFrame[26] > 0 //this didn't work btw
								ghost.eflags = $|MFE_FORCESUPER
							else
								ghost.eflags = $ & ~(MFE_FORCESUPER)
							end
							ghost.renderflags = curFrame[16]
							if ghost.state == S_PLAY_ROLL then
								if curFrame[17] > FRACUNIT*15
									local thokky = P_SpawnMobj(curFrame[1],curFrame[2],curFrame[3],MT_THOK)
									thokky.color = ghost.color
								end
							end
							if curFrame[19] ~= {}
								if frames[player.chrframe-1]
									local soundlist = curFrame[19]
									for sound = 1,#soundlist do
										local wasSoundPlayed = false
										local oldsoundlist = frames[player.chrframe-1][19]
										if oldsoundlist ~= {}
											for nsound = 1,#oldsoundlist do
												if oldsoundlist[nsound] == soundlist[sound]
													wasSoundPlayed = true
												end
											end
										end
										if not wasSoundPlayed
											local dummy = P_SpawnMobjFromMobj(ghost,0,0,0,MT_THOK)
											dummy.sprite = SPR_NULL
											dummy.state = ghost.state
											dummy.flags2 = $|MF2_DONTDRAW
											dummy.fuse = TICRATE*2
											S_StartSound(dummy,soundlist[sound])
										end
									end
								end
							end
							ghost.spriteyoffset = curFrame[27]
							ghost.spritexoffset = curFrame[28]
							ghost.spritexscale = curFrame[29]
							ghost.spriteyscale = curFrame[30]
							ghost.mirrored = curFrame[31]
							ghost.blendmode = curFrame[32]
							ghost.height = curFrame[33]
							ghost.radius = curFrame[34]
							
							if chrono_indicators.value == 1
								local indicator = P_SpawnMobjFromMobj(ghost,0,0,ghost.height + (FRACUNIT*5),MT_THOK)
								indicator.fuse = 1
								indicator.sprite = SPR_LCKN
								indicator.flags = $|MF_SCENERY
								indicator.frame = F|FF_FULLBRIGHT
								for i = 1,iteration do
									local bubblex = 0
									local bubbley = 0
									local bubblez = 0
									if i % 8 == 0
									or i % 8 == 4
										if i % 8 == 0
											bubbley = FRACUNIT*40
										else
											bubbley = -FRACUNIT*40
										end
									elseif i % 8 == 1
									or i % 8 == 3
										bubblex = FRACUNIT*20
										if i % 8 == 1
											bubbley = FRACUNIT*20
										else
											bubbley = -FRACUNIT*20
										end
									elseif i % 8 == 2
										bubblex = FRACUNIT*40
									elseif i % 8 == 5
									or i % 8 == 7
										bubblex = -FRACUNIT*20
										if i % 8 == 5
											bubbley = -FRACUNIT*20
										else
											bubbley = FRACUNIT*20
										end
									elseif i % 8 == 6
										bubblex = -FRACUNIT*40
									end
									bubblez = (i/9)
									local bubble = P_SpawnMobjFromMobj(indicator,bubblex,bubbley,9+(5*bubblez)*FRACUNIT*3,MT_THOK)
									bubble.fuse = 1
									bubble.flags = $|MF_SCENERY
									bubble.sprite = SPR_LCKN
									bubble.frame = G|FF_FULLBRIGHT
								end
							end
							
							if player.chrframe == table.maxn(frames)-1
								local dummy = P_SpawnMobjFromMobj(ghost,0,0,0,MT_THOK)
								dummy.sprite = SPR_NULL
								dummy.fuse = TICRATE*2
								S_StartSound(dummy,195)
								if chrono_blocks.value == 1
									P_SpawnMobjFromMobj(ghost,0,0,0,MT_CHRONOBLOCK)
								end
								for i = 0,359 do
									if i % 45 == 44
										local particle = P_SpawnMobjFromMobj(ghost,0,0,0,MT_PARTICLE)
										particle.angle = FixedAngle(i*FRACUNIT)
										particle.fuse = (TICRATE/3)*2
										P_InstaThrust(particle,particle.angle,FRACUNIT*5)
									end
								end
							end
							
							local remainingTime = table.maxn(frames)-player.chrframe
							if remainingTime < (TICRATE/2)+1
								if leveltime % 2 == 1
									local fading = P_SpawnGhostMobj(ghost)
									P_SetOrigin(fading,ghost.x,ghost.y,ghost.z)
									fading.blendmode = AST_ADD
									fading.fuse = (TICRATE/3)*2
									local particles = P_SpawnMobjFromMobj(ghost,FRACUNIT*P_RandomRange(-20,20),FRACUNIT*P_RandomRange(-20,20),FRACUNIT*P_RandomRange(0,20),MT_PARTICLE)
									P_SetObjectMomZ(particles,FRACUNIT*7,false)
									particles.blendmode = AST_ADD
									particles.fuse = TICRATE/2
								end
								ghost.frame = $|FF_ADD
							end
							
							if player.viewingIter == iteration
								player.powers[pw_nocontrol] = 1
								player.powers[pw_invulnerability] = 20
								P_TeleportCameraMove(camera,curFrame[20],curFrame[21],curFrame[22])
								camera.angle = curFrame[23]
								camera.chase = true
								camera.aiming = curFrame[24]
								if player.chrframe == #frames - 1
									player.viewingIter = nil
									P_KillMobj(player.mo)
								end
							end
							
							local shieldOrb
							
							if curFrame[18] ~= SH_NONE
								shieldOrb = P_SpawnMobjFromMobj(ghost,curFrame[1],curFrame[2],curFrame[3],MT_THOK)
								P_SetOrigin(shieldOrb,curFrame[1],curFrame[2],curFrame[3])
								shieldOrb.fuse = 1
								shieldOrb.scale = skins[curFrame[5]].shieldscale
								if curFrame[35] > 0
									shieldOrb.state = S_IVSP
								elseif curFrame[18] == SH_PITY
									shieldOrb.state = S_PITY1
								elseif curFrame[18] == SH_WHIRLWIND
									shieldOrb.state = S_WIND1
								elseif curFrame[18] == SH_ATTRACT
									shieldOrb.state = S_MAGN1
								elseif curFrame[18] & SH_FORCE
									shieldOrb.state = S_FORC1
								elseif curFrame[18] == SH_ARMAGEDDON
									shieldOrb.state = S_ARMB1
									local otherOrb = P_SpawnMobjFromMobj(ghost,curFrame[1],curFrame[2],curFrame[3],MT_THOK)
									P_SetOrigin(otherOrb,curFrame[1],curFrame[2],curFrame[3])
									otherOrb.fuse = 1
									otherOrb.scale = skins[curFrame[5]].shieldscale
									otherOrb.state = S_ARMA1
									local other2Orb = P_SpawnMobjFromMobj(ghost,curFrame[1],curFrame[2],curFrame[3],MT_THOK)
									P_SetOrigin(other2Orb,curFrame[1],curFrame[2],curFrame[3])
									other2Orb.fuse = 1
									other2Orb.scale = skins[curFrame[5]].shieldscale
									other2Orb.state = S_ARMF1
								elseif curFrame[18] == SH_ELEMENTAL
									shieldOrb.state = S_ELEM1
									local otherOrb = P_SpawnMobjFromMobj(ghost,curFrame[1],curFrame[2],curFrame[3],MT_THOK)
									P_SetOrigin(otherOrb,curFrame[1],curFrame[2],curFrame[3])
									otherOrb.fuse = 1
									otherOrb.scale = skins[curFrame[5]].shieldscale
									otherOrb.state = S_ELEMF1
								elseif curFrame[18] == SH_FLAMEAURA
									shieldOrb.state = S_FIRSB1
									local otherOrb = P_SpawnMobjFromMobj(ghost,curFrame[1],curFrame[2],curFrame[3],MT_THOK)
									P_SetOrigin(otherOrb,curFrame[1],curFrame[2],curFrame[3])
									otherOrb.fuse = 1
									otherOrb.scale = skins[curFrame[5]].shieldscale
									otherOrb.state = S_FIRS1
								elseif curFrame[18] == SH_BUBBLEWRAP
									shieldOrb.state = S_BUBSB1
									local otherOrb = P_SpawnMobjFromMobj(ghost,curFrame[1],curFrame[2],curFrame[3],MT_THOK)
									P_SetOrigin(otherOrb,curFrame[1],curFrame[2],curFrame[3])
									otherOrb.fuse = 1
									otherOrb.scale = skins[curFrame[5]].shieldscale
									otherOrb.state = S_BUBS1
								elseif curFrame[18] == SH_THUNDERCOIN
									shieldOrb.state = S_ZAPSB1
									local otherOrb = P_SpawnMobjFromMobj(ghost,curFrame[1],curFrame[2],curFrame[3],MT_THOK)
									P_SetOrigin(otherOrb,curFrame[1],curFrame[2],curFrame[3])
									otherOrb.fuse = 1
									otherOrb.scale = skins[curFrame[5]].shieldscale
									otherOrb.state = S_ZAPS1
								end
							end
						end
					end
				end
			end
			
			if chrono_timescale.value == 0 then
				--player.chrframe = $ + 0
			elseif chrono_timescale.value == 2 then
				if leveltime % 10 == 9
					player.chrframe = $ + 1
				end
			elseif chrono_timescale.value == 1 then
				if leveltime % 19 == 18
					player.chrframe = $ + 1
				end
			elseif chrono_timescale.value == 3 then
				player.chrframe = $ + 1
			elseif chrono_timescale.value == 4 then
				player.chrframe = $ + 2
			elseif chrono_timescale.value == 5 then
				player.chrframe = $ + 3
			end
			
			
			if player.retryHold == nil
				player.retryHold = 0
			end
			if player.cmd.buttons & BT_TOSSFLAG
			and player.powers[pw_carry] ~= CR_NIGHTSMODE
			and player.powers[pw_carry] ~= CR_NIGHTSFALL
				player.retryHold = $ + 1
			else
				player.retryHold = 0
			end
			
			
			//Disabling retrying through pause menu
			if player.mo.health
				player.lives = 1
			else
				player.lives = INFLIVES
			end
			
			
			if chrono_timescale.value == 0 then
				player.curTime = $ + 0
			elseif chrono_timescale.value == 2 then
				if leveltime % 10 == 9
					player.curTime = $ + 1
				end
			elseif chrono_timescale.value == 1 then
				if leveltime % 19 == 18
					player.curTime = $ + 1
				end
			elseif chrono_timescale.value == 3 then
				player.curTime = $ + 1
			elseif chrono_timescale.value == 4 then
				player.curTime = $ + 2
			elseif chrono_timescale.value == 5 then
				player.curTime = $ + 3
			end
			
			if chrono_timeline.value == 1
			or player.powers[pw_carry] == CR_NIGHTSMODE
			or player.powers[pw_carry] == CR_NIGHTSFALL
				hud.disable("lives")
			else
				hud.enable("lives")
			end
			
			if player.retryHold > TICRATE+(TICRATE/2)
				if player.mo.health
					P_KillMobj(player.mo,nil,nil)
				end
			end
			
			if chrono_timescale.value != 3
				player.invalidatedIter = true
			end
			
			if player.invalidatedIter == true
				player.iterations[player.iteration] = {}
			end
		end
	end
end)

addHook("MobjDeath", function(mo,inflictor,source,damagetype)
	if mo.type == MT_PLAYER
		mo.player.viewingIter = nil
		if not mo.player.invalidatedIter
			if mo.player.retryHold > TICRATE+(TICRATE/2)-1
				addToPlayerTimeline(mo.player,mo.player.curTime,{"T_ITERATETFLG",mo.player.iteration,damagetype})
			else
				addToPlayerTimeline(mo.player,mo.player.curTime,{"T_ITERATE",mo.player.iteration,damagetype,P_IsObjectInGoop(mo),P_InQuicksand(mo),P_InSpaceSector(mo),mo.eflags})
			end
		end
		mo.player.suffocatingInWater = nil
		mo.player.hasalreadyturnedsuper = nil
	end
end,MT_PLAYER)

addHook("PlayerSpawn", function(player)
	if player.mo
		if player.invalidatedIter or player.origMap ~= gamemap
			player.invalidatedIter = nil
			player.origMap = gamemap
		else
			player.iteration = $ + 1
		end
		
		//For anyone wondering about why you can't view iterations that start from checkpoints, this is why
		//
		//We have to sync curTime with the time it was on the timeline when you got the starpost
		//in order to prevent desynchronization, but that causes some more issues involving the
		//playback and recording systems.
		//
		//If we don't update chrframe, now the iterations themselves are desynced with the timeline.
		//If we don't update recframe, recording would start at frame 1 even though you're not on frame 1,
		//causing them to play several seconds or minutes behind and to desync with the timeline.
		//
		//But that would mean the frames between 0 and the updated recframe are empty, so the replay code
		//freaks out when we try to relive that iteration.
		//
		//There's not a fix to this - at least, not that I know of. So while I'm figuring that out,
		//reliving starpost-starting iterations is disabled.
		if player.starpostCurTime
			player.curTime = player.starpostCurTime
			player.chrframe = player.curTime
			player.recframe = player.curTime
		else
			player.curTime = 1
			player.chrframe = 1
			player.recframe = 1
		end
		
		//Checking freeslots for new sounds
		for i = 736,2327 do
			if sfxinfo[i]
				if (string.sub(sfxinfo[i].name,0,2) == "fr" and string.upper(string.sub(sfxinfo[i].name,3)) == string.sub(sfxinfo[i].name,3))
				or (string.sub(sfxinfo[i].name,0,3) == "fre" and string.upper(string.sub(sfxinfo[i].name,4)) == string.sub(sfxinfo[i].name,4))
				or (string.sub(sfxinfo[i].name,0,4) == "free" and string.upper(string.sub(sfxinfo[i].name,5)) == string.sub(sfxinfo[i].name,5))
					continue
				end
				local isOnList = false
				for f = 1,#chronoSoundlist do
					if chronoSoundlist[f] == i
						isOnList = true
					end
				end
				if not isOnList
					rawset(chronoSoundlist,#chronoSoundlist+1,i)
				else
					
				end
			else
				break
			end
		end
	end
end)

addHook("MapChange", function(newmap)
	if newmap ~= gamemap
		for player in players.iterate do
			player.recframe = 1
			player.cbrframe = 1
			player.iterations = {}
			player.iteration = 1
			player.invalidatedIter = nil
			player.timeline = {}
			player.curTime = 1
			player.starpostCurTime = nil
			addToPlayerTimeline(player,1,{"T_START"})
			addToPlayerTimeline(player,TICRATE*60,{"T_1MIN"})
			addToPlayerTimeline(player,TICRATE*120,{"T_2MIN"})
			addToPlayerTimeline(player,TICRATE*180,{"T_3MIN"})
			addToPlayerTimeline(player,TICRATE*240,{"T_4MIN"})
			addToPlayerTimeline(player,TICRATE*300,{"T_5MIN"})
			addToPlayerTimeline(player,TICRATE*360,{"T_6MIN"})
			addToPlayerTimeline(player,TICRATE*420,{"T_7MIN"})
			addToPlayerTimeline(player,TICRATE*480,{"T_8MIN"})
			addToPlayerTimeline(player,TICRATE*540,{"T_9MIN"})
			addToPlayerTimeline(player,TICRATE*600,{"T_10MIN"})
		end
	end
end)

addHook("TouchSpecial",function(mo,toucher)
	if mo.type == MT_STARPOST and mo.state == S_STARPOST_IDLE
		if toucher.type == MT_PLAYER
			toucher.player.starpostCurTime = toucher.player.curTime
			addToPlayerTimeline(toucher.player,toucher.player.curTime,{"T_CHECKPOINT",toucher.player.iteration,toucher.player.starpostnum})
			return false
		end
	end
end,MT_STARPOST)

addHook("TouchSpecial",function(mo,toucher)
	if mo.type == MT_EXTRALARGEBUBBLE and mo.state == S_EXTRALARGEBUBBLE
		if toucher.type == MT_PLAYER
			if toucher.player.powers[pw_underwater] < 403
				addToPlayerTimeline(toucher.player,toucher.player.curTime,{"T_BUBBLE",toucher.player.iteration})
				toucher.player.suffocatingInWater = nil
				return false
			end
		end
	end
end,MT_EXTRALARGEBUBBLE)

/*COM_AddCommand("chrono_saveiter",function(player,iteration)
	if not iteration then
		CONS_Printf(player,"chrono_saveiter - Save a previous iteration.")
		if not player.iterations
			CONS_Printf(player,"Doesn't seem like there are any iterations here.")
		elseif player.iteration == 1
			CONS_Printf(player,"The only iteration is the one you're currently recording.")
		else
			CONS_Printf(player,"Following iterations are available to save (properties are from frame 1):")
			local iterCount = player.iterations
			for i = 1,#iterCount do
				if i ~= player.iteration
					CONS_Printf(player,"\x85       Iteration " .. i .. " - \x84" .. R_GetNameByColor(iterCount[i][1][6]) .. " " .. iterCount[i][1][5])
				end
			end
		end
	else
		local iterCount = player.iterations
		iteration = tonumber(iteration)
		if iterCount
			if iteration == player.iteration
				CONS_Printf(player,"\x85 You need to finish this iteration before you can save it.")
			elseif iteration > #iterCount or iteration < 1
				CONS_Printf(player,"\x85 Invalid iteration.")
			else
				local id = P_RandomRange(10000,65534)
				local file = io.openlocal("chronotron/" .. gamemap .. "/" .. iterCount[iteration][1][5] .. "/iter" .. iteration .. "_" .. id .. ".txt","a+")
				local selIter = iterCount[iteration]
				for frame = 1,#selIter do
					local frameProp = selIter[frame]
					for property = 1,#frameProp do
						file:write(tostring(frameProp[property]) .. ", ")
					end
					file:write("\n")
				end
				file:flush()
				file.close(file)
				CONS_Printf(player,"\x83 Iteration saved as iter" .. iteration .. "_" .. id .. ".txt")
			end
		end
	end
end,COM_LOCAL)*/

COM_AddCommand("chrono_relive",function(player,iteration)
	if not iteration then
		CONS_Printf(player,"chrono_relive - View a previous iteration.")
		if not player.iterations
			CONS_Printf(player,"Doesn't seem like there are any iterations here.")
		elseif player.iteration == 1
			CONS_Printf(player,"The only iteration is the one you're currently recording.")
		else
			CONS_Printf(player,"Following iterations are available to view (properties are from frame 1):")
			local iterCount = player.iterations
			for i = 1,#iterCount do
				if i ~= player.iteration and iterCount[i][1] ~= nil
					if iterCount[i][1][25] == 0
						CONS_Printf(player,"\x85       Iteration " .. i .. " - \x84" .. R_GetNameByColor(iterCount[i][1][6]) .. " " .. iterCount[i][1][5])
					end
				end
			end
		end
	else
		local iterCount = player.iterations
		iteration = tonumber(iteration)
		if iterCount
			if iteration < 1
			or iteration > #iterCount
				CONS_Printf(player,"\x85 Iteration " .. iteration .. " doesn't exist.")
			elseif not player.mo
				CONS_Printf(player,"\x85 You need to be in a level to do this.")
			elseif (not iterCount[iteration][1]) or (iterCount[iteration][1][25] > 0)
				CONS_Printf(player,"\x85 Iterations that start from checkpoints cannot be relived.")
			else
				player.chrframe = 1
				player.curTime = 1
				player.recframe = 1
				player.invalidatedIter = true
				player.viewingIter = iteration
			end
		end
	end
end,COM_LOCAL)

hud.add(function(v,plr,cam)
	if plr.mo
		if plr.iteration
			if plr.invalidatedIter
				v.drawString(16,166,"Invalid Iteration",V_ALLOWLOWERCASE|V_REDMAP|V_PERPLAYER|V_HUDTRANS|V_SNAPTOLEFT|V_SNAPTOBOTTOM)
			elseif plr.powers[pw_carry] == CR_NIGHTSMODE
			or plr.powers[pw_carry] == CR_NIGHTSFALL
				v.drawString(16,166,"Chronotron Inactive...",V_ALLOWLOWERCASE|V_REDMAP|V_PERPLAYER|V_HUDTRANS|V_SNAPTOLEFT|V_SNAPTOBOTTOM)
			else
				v.drawString(16,166,"Iteration " .. plr.iteration,V_ALLOWLOWERCASE|V_GREENMAP|V_PERPLAYER|V_HUDTRANS|V_SNAPTOLEFT|V_SNAPTOBOTTOM)
			end
		else
			v.drawString(16,166,"Chronotron Booting...",V_ALLOWLOWERCASE|V_REDMAP|V_PERPLAYER|V_HUDTRANS|V_SNAPTOLEFT|V_SNAPTOBOTTOM)
		end
		
		
		//SELF-DESTRUCT
		if plr.retryHold
			if plr.retryHold > 0 and plr.mo.health
				v.drawString(160,100,"SELF-DESTRUCT",V_BLUEMAP|V_PERPLAYER,"center")
				v.drawString(160,110,G_TicsToSeconds((TICRATE+(TICRATE/2))-plr.retryHold) .. "." .. G_TicsToCentiseconds((TICRATE+(TICRATE/2))-plr.retryHold),V_BLUEMAP|V_PERPLAYER,"center")
			end
		end
		
		
		//Timeline
		if chrono_timeline.value == 1
		and plr.powers[pw_carry] ~= CR_NIGHTSMODE
		and plr.powers[pw_carry] ~= CR_NIGHTSFALL
			v.drawScaled(FRACUNIT*155,FRACUNIT*186,FRACUNIT,v.cachePatch("TIMELINE"),V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_HUDTRANS)
			
			if plr.timeText
				v.drawString(300,166,plr.timeText,V_ALLOWLOWERCASE|V_PERPLAYER|V_HUDTRANS|V_SNAPTOLEFT|V_SNAPTOBOTTOM,"thin-right")
			end
			
			if plr.timeline
				local timeline = plr.timeline
				local maxTime = 0
				for eventTime,eventProp in pairs(timeline) do
					local posx = FRACUNIT*(155-(plr.curTime-eventTime))
					if posx > TIMELINE_MIN and posx < TIMELINE_MAX
						v.drawScaled(posx,FRACUNIT*186,FRACUNIT,v.cachePatch(eventProp[1]),V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_HUDTRANS)
					end
				end
			end
		end
	end
end)


addHook("MobjThinker",function(mo)
	if mo.type == MT_CHRONOBLOCK
		local victim = P_LookForPlayers(mo,FRACUNIT*20,true,false)
		if victim then
			if mo.target.health
			and mo.z <= (mo.target.z + mo.target.height)
			and mo.z > (mo.target.z - FRACUNIT)
				P_KillMobj(mo.target,mo,nil,DMG_CRUSHED)
			end
			mo.target = nil
		end
	end
end)